/*
 * Include file:
 *	main.h
 */

/*
 * Globally available externs
 */
extern unsigned int	no_of_files;
extern char             *current_filename;
extern short		pass;
extern unsigned int	file_number;
extern int		page_length;
extern int		page_width;
extern long		total_pages;
extern boolean		use_environment;
