/*
 * Include file:
 *	lang_c.h
 */
extern stream_status	get_c_char(char *, char_status *);
extern char		lang_c_defaults[];
extern stream_status	get_pc_char(char *, char_status *);
extern char		lang_pc_defaults[];
